/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java_cup.runtime.Symbol;

%%


%{
    // Max size of string constants
    static int MAX_STR_CONST = 1024;

    // For assembling string constants
    StringBuffer string_buf = new StringBuffer();

    // For line numbers
    private int curr_lineno = 1;
    int get_curr_lineno() {
	return curr_lineno;
    }

    private String filename;

    void set_filename(String fname) {
      filename = fname;
    }

    String curr_filename() {
	return filename;
    }

    /* For counting open (*'s in nested comments */
    private int count_open_com = 0;

    /* This flag is to be used by the code executed when EOF occurs */
    private int EOF_flag = 0;

    /* This method helps handle string constants  that are too long by 
     * reinitializing the string buffer, changing the lexical state and 
     * returning an error Symbol.  
     */
     Symbol string_too_long(){
           string_buf = new StringBuffer(); 
	    yybegin (STRING_ERROR);
           return new Symbol(BLOGTokenConstants.ERROR,
			      "String constant too long"); 
     }
     
    
%}



%init{
    // empty for now
%init}


%eofval{
    switch(yy_lexical_state) {
    case YYINITIAL:
	/* nothing special to do in the initial state */
	break;
    case PAREN_COMMENT:
      System.out.println("Error: EOF inside comment, line "+ curr_lineno +".");
      System.exit(1);
	
    }
    /* Reinitialize everything before signaling EOF */
    EOF_flag = 0;
    string_buf = new StringBuffer();
    yybegin(YYINITIAL);
    count_open_com = 0;
    curr_lineno = 1;
    return new Symbol(BLOGTokenConstants.EOF);
%eofval}

%class BLOGLexer
%cup


VTAB            = \x0b


NEWLINE		= \n


WHITESPACE	= [ \f\r\t]|VTAB 

%state LINE_COMMENT, PAREN_COMMENT, QUOTED, BACKSL_IN_QUOT, STRING_ERROR, STRING_ERROR_BACKSL

%%

<YYINITIAL>{NEWLINE}	 { curr_lineno++; } 
<YYINITIAL>{WHITESPACE} { /* Do nothing */}
<YYINITIAL>"/*" { yybegin (PAREN_COMMENT); }

<PAREN_COMMENT>"*/" {  string_buf = new StringBuffer();
                       yybegin(YYINITIAL); }
<PAREN_COMMENT>{NEWLINE} { curr_lineno++; }
<PAREN_COMMENT>. {}

<YYINITIAL>"//" {yybegin (LINE_COMMENT); }
<LINE_COMMENT>{NEWLINE} { curr_lineno++;
                          yybegin (YYINITIAL); } 
<LINE_COMMENT>. {}
<YYINITIAL>"\"" { /* Open-quote. Start collecting a string constant*/  yybegin (QUOTED);}
<QUOTED>"\"" { /* Close-quote seen */
       Symbol s =   new Symbol(BLOGTokenConstants.STR_CONST, 
			                            string_buf.toString());
       string_buf = new StringBuffer(); /* reinitialize the buffer */
       yybegin(YYINITIAL);
       return s;}
				
 
<QUOTED>{NEWLINE}  { curr_lineno++; 
                     yybegin (YYINITIAL);
		     string_buf = new StringBuffer(); /* reinit the buffer */
                     return new Symbol(BLOGTokenConstants.ERROR, 
                                    "Unterminated string constant"); }
<QUOTED>\0 {  string_buf = new StringBuffer(); /* null in the string */
	        yybegin (STRING_ERROR);
                return new Symbol(BLOGTokenConstants.ERROR,  
				  "String contains null character"); }


<STRING_ERROR>{NEWLINE} {curr_lineno++; }
<STRING_ERROR>[\\]  {yybegin(STRING_ERROR_BACKSL);}
<STRING_ERROR_BACKSL>{NEWLINE}  {curr_lineno++; yybegin(STRING_ERROR);}
<STRING_ERROR_BACKSL>.|{WHITESPACE}  { 
  /* to ignore escaped quotes etc */ yybegin(STRING_ERROR);}
<STRING_ERROR>"\"" { yybegin(YYINITIAL);}
<STRING_ERROR>.|{WHITESPACE}   { /* ignore the rest  */ }

<QUOTED>[\\]  { /* Backslash seen in string */  yybegin(BACKSL_IN_QUOT); }
<BACKSL_IN_QUOT>"\"" { yybegin (QUOTED); 
      if (string_buf.length() < MAX_STR_CONST) string_buf.append('\"');
      else return string_too_long();}
<BACKSL_IN_QUOT>[\\]  {/* second backslash seen */  yybegin (QUOTED); 
 if (string_buf.length() < MAX_STR_CONST) string_buf.append('\\'); 
      else return string_too_long();}
<BACKSL_IN_QUOT>n  {  yybegin (QUOTED); 
 if (string_buf.length() < MAX_STR_CONST) string_buf.append('\n');
      else return string_too_long(); }
<BACKSL_IN_QUOT>{NEWLINE}  { curr_lineno++; yybegin (QUOTED); 
  if (string_buf.length() < MAX_STR_CONST) string_buf.append('\n'); 
      else return string_too_long();}
<BACKSL_IN_QUOT>t { yybegin (QUOTED);
 if (string_buf.length() < MAX_STR_CONST) string_buf.append('\t');
     else return string_too_long(); }
<BACKSL_IN_QUOT>b { yybegin (QUOTED);
 if (string_buf.length() < MAX_STR_CONST) string_buf.append('\b'); 
     else return string_too_long();}
<BACKSL_IN_QUOT>f { yybegin (QUOTED); 
 if (string_buf.length() < MAX_STR_CONST) string_buf.append('\f');
     else return string_too_long(); }
<BACKSL_IN_QUOT>.|{WHITESPACE} {/*a backslash followed by regular char is 
				                                     ignored */
                 yybegin (QUOTED); 
                 if (string_buf.length() < MAX_STR_CONST) 
		   string_buf.append(yytext());
                 else 
		   return string_too_long();}


<QUOTED>.|{WHITESPACE} { 
 if (string_buf.length() < MAX_STR_CONST) string_buf.append(yytext());
     else return string_too_long();}



<YYINITIAL>-?[0-9]*[0-9][.][0-9][0-9]* { return 
				    new Symbol(BLOGTokenConstants.DOUBLE_CONST,
					                           yytext()); }

<YYINITIAL>-?[0-9][0-9]*  { /* Integers */
                          return new Symbol(BLOGTokenConstants.INT_CONST,
					                           yytext()); }

<YYINITIAL>[Tt][Yy][Pp][Ee]     { return new Symbol(BLOGTokenConstants.TYPE); }
<YYINITIAL>[Rr][Aa][Nn][Dd][Oo][Mm] { return 
				   new Symbol(BLOGTokenConstants.RANDOM); }
<YYINITIAL>[Nn][Oo][Nn][Rr][Aa][Nn][Dd][Oo][Mm] { return 
				   new Symbol(BLOGTokenConstants.NONRANDOM); }
<YYINITIAL>[Gg][Ee][Nn][Ee][Rr][Aa][Tt][Ii][Nn][Gg] { return 
				   new Symbol(BLOGTokenConstants.GENERATING); }
<YYINITIAL>[Gg][Uu][Aa][Rr][Aa][Nn][Tt][Ee][Ee][Dd] { return 
				   new Symbol(BLOGTokenConstants.GUARANTEED); }


<YYINITIAL>[Tt][Hh][Ee][Nn]     { return new Symbol(BLOGTokenConstants.THEN); }
<YYINITIAL>[Ee][Ll][Ss][Ee]  	{ return new Symbol(BLOGTokenConstants.ELSE); }
<YYINITIAL>[Ff][Oo][Rr]         { return new Symbol(BLOGTokenConstants.FOR); }
<YYINITIAL>f[Aa][Ll][Ss][Ee]   { return new Symbol(BLOGTokenConstants.FALSE); }

<YYINITIAL>[Ee][Ll][Ss][Ee][Ii][Ff]  { return 
				       new Symbol(BLOGTokenConstants.ELSEIF); }
<YYINITIAL>[Ii][Ff]  		{ return new Symbol(BLOGTokenConstants.IF); }

<YYINITIAL>t[Rr][Uu][Ee]	{ return new Symbol(BLOGTokenConstants.TRUE); }
<YYINITIAL>[Qq][Uu][Ee][Rr][Yy]	{ return new Symbol(BLOGTokenConstants.QUERY);}
<YYINITIAL>[Oo][Bb][Ss]         { return new Symbol(BLOGTokenConstants.OBS);}
<YYINITIAL>"="			{ return new Symbol(BLOGTokenConstants.EQ); }
<YYINITIAL>"&"			{ return new Symbol(BLOGTokenConstants.AND); }
<YYINITIAL>"|"                  { return new Symbol(BLOGTokenConstants.OR); }
<YYINITIAL>"~"			{ return 
				    new Symbol(BLOGTokenConstants.DISTRIB); }
<YYINITIAL>"!"			{ return new Symbol(BLOGTokenConstants.NEG); }
<YYINITIAL>","			{ return new Symbol(BLOGTokenConstants.COMMA);}
<YYINITIAL>"->"			{ return 
				    new Symbol(BLOGTokenConstants.RIGHTARROW);}
<YYINITIAL>";"			{ return new Symbol(BLOGTokenConstants.SEMI); }
<YYINITIAL>":"			{ return new Symbol(BLOGTokenConstants.COLON);}
<YYINITIAL>"("			{ return 
				    new Symbol(BLOGTokenConstants.LPAREN); }
<YYINITIAL>")"			{ return 
				    new Symbol(BLOGTokenConstants.RPAREN); }
<YYINITIAL>"}"			{ return 
				    new Symbol(BLOGTokenConstants.RBRACE); }
<YYINITIAL>"{"			{ return 
				    new Symbol(BLOGTokenConstants.LBRACE); }
<YYINITIAL>"["			{ return 
				    new Symbol(BLOGTokenConstants.LBRACKET); }
<YYINITIAL>"]"			{ return 
				    new Symbol(BLOGTokenConstants.RBRACKET); }
<YYINITIAL>"#"                  { return 
                                    new Symbol(BLOGTokenConstants.NUMSIGN); } 

<YYINITIAL>[Ee][Xx][Ii][Ss][Tt][Ss] { /* existential quantifier */
	return new Symbol(BLOGTokenConstants.EXISTS); }
<YYINITIAL>[Ff][Oo][Rr][Aa][Ll][Ll] { /* universal quantifier */
	return new Symbol(BLOGTokenConstants.FORALL); }
<YYINITIAL>[A-Za-z][A-Za-z0-9_]*  {
        return new Symbol(BLOGTokenConstants.ID, yytext()); }
<YYINITIAL>[A-Za-z][A-Za-z0-9_]*([.][A-Za-z][A-Za-z0-9_]*)* {
        return new Symbol(BLOGTokenConstants.CLASS_NAME, yytext()); }
<YYINITIAL>.                     { return new Symbol(BLOGTokenConstants.ERROR, 
                                          yytext()); }


.             { /*
                    *  This should be the very last rule and will match
                    *  everything not matched by other lexical rules.
                    */
                   System.err.println("LEXER BUG - UNMATCHED: " + yytext()); }























