/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;


/** Represents a single a variable.
 *
 * @see blog.Term
 */
public class VariableTerm extends Term {


    public VariableTerm( String name ){

	this.name = name;
  
    }
 

    public String getVarName( ){

	return name;

    }

    /**
     * Returns the non-random value of this term if it is non-random; 
     * otherwise returns null.  
     *
     * @param a the Assignment to use for evaluating variables in this term
     */
    public Object getValueIfNonRandom(Assignment a) {
	return a.getVal(name);
    }

    /**
     * Returns the value of this variable under the given assignment, 
     * and an empty parent set.  
     */
    public ParentsAndValue getParentsAndValue(PartialWorld w, 
					      Assignment a, 
					      ValueChooser chooser) {
	return new ParentsAndValue(Collections.EMPTY_SET, a.getVal(name));
    }

    /**
     * Returns a new variable term whose name is obtained by looking 
     * up this term's variable name in the given renaming map.  If this 
     * term's variable name is not in the given map, then the new 
     * term has the same variable name as this term.  
     *
     * @param renaming a Map from String to String where the keys are old 
     *                 variable names and the values are new variable names
     */
    public Term getFormWithVarsRenamed(Map renaming) {
	String newName = name;
	if (renaming.containsKey(name)) {
	    newName = (String) renaming.get(name);
	}
	return new VariableTerm(newName);
    }

    public Set getVariables() {
	return Collections.singleton(name);
    }

    public boolean containsTerm(Term target) {
	return equals(target);
    }

    public Set getGenFuncsApplied(Term subject) {
	return Collections.EMPTY_SET;
    }

    /**
     * Two VariableTerms are equal if they have the same name.
     */
    public boolean equals(Object o) {
	if (o instanceof VariableTerm) {
	    VariableTerm other = (VariableTerm) o;
	    return (name.equals(other.getVarName()));
	}
	return false;
    }

    public int hashCode() {
	return name.hashCode();
    }

    /**
     * Returns the name of this variable.
     */
    public String toString() {
	return name;
    }

    public boolean checkTypesAndScope(Map scope) {
	if (scope.containsKey(name)) {
	    return true;
	}
	
	typeError("Variable \"" + name + "\" not in scope");
	return false;
    } 

    public Type getType(Map scope) {
	return (Type) scope.get(name);
    }

    private String name;

}
